# Copyright (C) 2011 Oracle. All rights reserved.
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

'''Some basic monitor data consumers and helpers'''

from twisted.application.service import Service
from twisted.internet.task import LoopingCall
from twisted.python import log

from zope.interface import implements

from monitor.iconsumer import IConsumer
from monitor.service import EventItem

#pylint: disable-msg=C0103

class SimpleConsumer(Service):
    '''A simple data consumer, sending all data to a callable'''
    implements(IConsumer)

    def __init__(self, publish_fun):
        '''Initialize a new SimpleConsumer

        @param publish_fun: Function to call when data should be consumed
        @type publish_fun: callable
        '''
        self.publish_fun = publish_fun

    def publish(self, data):
        '''Publish a data item'''
        self.publish_fun(data)


class LogMsgConsumer(SimpleConsumer):
    '''
    Simple data consumer which logs all data using L{twisted.python.log.msg}
    '''
    def __init__(self, system='monitor.logconsumer'):
        '''Initialize a new LogMsgConsumer

        @param system: Log system name
        @type system: str
        '''
        SimpleConsumer.__init__(self,
                                lambda d: log.msg(str(d), system=system))


class BufferedConsumer(Service):
    '''A consumer which buffers data and sends it to a real consumer in bulk

    The real consumer should subclass this class and implement publish_list,
    which takes one argument: an iterable of all data to consume.
    '''
    implements(IConsumer)

    def __init__(self, timeout, buffer_events=False):
        '''Initialize a new BufferedConsumer

        @param timeout: Timeout (in seconds) between buffer flushes
        @type timeout: number
        @param buffer_events: Buffer event messages
        @type buffer_events: bool
        '''
        self.timeout = timeout
        self.buffer_events = buffer_events
        self.queue = list()
        self.task = LoopingCall(self.flush)

    def publish(self, data):
        '''Publish new data

        This adds new data to the buffer/queue.
        '''
        if not self.buffer_events and isinstance(data, EventItem):
            self.publish_list((data, ))
        else:
            self.queue.append(data)

    def startService(self):
        '''Start the service

        This sets up the flush loop.
        '''
        self.task.start(self.timeout)
        return Service.startService(self)

    def stopService(self):
        '''Stop the service

        This stops the flush loop.
        '''
        self.task.stop()
        self.flush()
        return Service.stopService(self)

    def flush(self):
        '''Flush the queue to C{publish_list}'''
        data = self.queue
        self.queue = list()
        if data:
            self.publish_list(data)

    def publish_list(self, data):
        '''Publish all items found in C{data}

        @param data: List of data items to process
        @type data: iterable
        '''
        raise NotImplementedError
