#!/usr/bin/env nosetests

# Copyright (C) 2011 Oracle. All rights reserved.
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

'''Monitor data consumer storing all items on a filesystem'''

import os
import time
import os.path
import stat
import tempfile

try:
    import functools
except ImportError:
    # Python 2.4 or lower...
    import types

    functools = types.ModuleType('functools')

    # Some stubs we need
    functools.wraps = lambda _: lambda f: f

    def partial(fun, *args, **kwargs):
        @functools.wraps(fun)
        def wrapped(*fargs, **fkwargs):
            newkwargs = kwargs.copy()
            newkwargs.update(fkwargs)
            return fun(*(args + fargs), **newkwargs)

        return wrapped

    functools.partial = partial

    del types, partial


try:
    import json
except ImportError:
    import simplejson as json

from twisted.python import log

try:
    from nose.tools import nottest
except ImportError:
    def nottest(func):
        '''Mark a function as not being a test'''

        func.__test__ = False

        return func


from monitor.consumers.basic import BufferedConsumer
from monitor.service import MonitorDataItem, EventItem

#pylint: disable-msg=C0103
msg = functools.partial(log.msg, system='monitor.consumer.filesystem')
#pylint: enable-msg=C0103

SERIALIZATION_VERSION = 1
TEMP_NAME = 'tmp'

def nofail(fun):
    '''Wrap a function so it never raises any exceptions

    Exceptions are logged using Twisted's L{twisted.python.log.err} function.

    Note: the wrapped function will not return any values, this decorator
    should be used on void functions only.

    Example:

    >>> def f():
    ...     print 'I\\'m about to blow...'
    ...     raise Exception('Boom')

    >>> g = nofail(f)
    >>> # Note this will show up in the console when usine Nosetests due to
    >>> # nofail using twisted.python.log.err, you can ignore this.
    >>> g()
    I'm about to blow...

    @param fun: Function to wrap
    @type fun: `callable`
    '''

    #pylint: disable-msg=C0111
    @functools.wraps(fun)
    def _wrapped(*args, **kwargs):
        try:
            fun(*args, **kwargs)
        except Exception, exc: #pylint: disable-msg=W0703
            log.err(exc, 'Failure in nofail function')

    return _wrapped


# Testcase-related utility functions
@nottest
def tempdir_test(fun):
    '''Wrapper for test functions which require a temporary working folder'''

    import shutil

    @functools.wraps(fun)
    def _wrapped(): #pylint: disable-msg=C0111
        dir_ = tempfile.mkdtemp()

        try:
            fun(dir_)
        finally:
            shutil.rmtree(dir_)

    return _wrapped


class FilesystemConsumer(BufferedConsumer):
    '''Monitor data consumer storing collected data on a filesystem'''

    # FIXME: we only want to retain the latest 30 minutes events and stats:
    #  30 * 60 / 20 = 90
    tempfile_max = 90
    event_tempfile_count = 0
    stats_tempfile_count = 0

    def __init__(self, stats_location, events_location, buffer_timeout=5,
                    buffer_events=False):
        '''Initialize a new C{FilesystemConsumer}

        @param stats_location: Filesystem-location for stats storage
        @type stats_location: `str`
        @param events_location: Filesystem-location for event storage
        @type events_location: `str`
        @param buffer_timeout: Buffer flush timeout (in seconds)
        @type buffer_timeout: number
        @param buffer_events: Buffer events
        @type buffer_events: `bool`
        '''

        msg('Creating filesystem consumer')

        # Make sure the stats and event container folders exist and are
        # writable
        check_folder(stats_location)
        check_folder(events_location)

        self._stats_location = os.path.abspath(stats_location)
        self._events_location = os.path.abspath(events_location)

        # Create temp folders
        self._stats_temp = create_temp(self._stats_location)
        self._events_temp = create_temp(self._events_location)

        BufferedConsumer.__init__(self, buffer_timeout, buffer_events)

    def publish_list(self, data):
        '''Publish a list of monitor items

        This is required by the abstract L{BufferedConsumer} class.

        @param data: Iterable of monitor data items
        @type data: iterable
        '''

        items = tuple(data)

        stats = tuple(item for item in items
                      if isinstance(item, MonitorDataItem))
        events = tuple(item for item in items
                       if isinstance(item, EventItem))

        if len(stats) + len(events) != len(items):
            msg('Received item(s) of unknown type, not all data will be ' \
                'processed')

        map(self._process_event, events) #pylint: disable-msg=W0141
        self._process_stats(stats)

    @nofail
    def _process_event(self, event):
        '''Process one event

        @param event: Event to process
        @type event: C{EventItem}
        '''

        serialized_event = serialize_event(event)
        tmp_name = write_temp(serialized_event, self._events_temp, str(self.event_tempfile_count))
        self.event_tempfile_count = (self.event_tempfile_count + 1) % self.tempfile_max
        relink(tmp_name, self._events_location)

    @nofail
    def _process_stats(self, stats):
        '''Process statistics values

        @param stats: Statistics to process
        @type stats: iterable
        '''

        stat_dicts = list()

        stat_dict = None
        for stat_ in stats:
            try:
                stat_dict = dump_stat(stat_)
            except Exception, exc: #pylint: disable-msg=W0703
                log.err(exc, 'Unable dump stat value')
                stat_dict = None
            else:
                if stat_dict:
                    stat_dicts.append(stat_dict)

        try:
            serialized_stats = serialize_stats(stat_dicts)
        except Exception, exc: #pylint: disable-msg=W0703
            log.err(exc, 'Unable to serialize stats')
        else:
            tmp_name = write_temp(serialized_stats, self._stats_temp, str(self.stats_tempfile_count))
            self.stats_tempfile_count = (self.stats_tempfile_count + 1) % self.tempfile_max
            relink(tmp_name, self._stats_location)

@tempdir_test
def test_filesystem_consumer(base):
    '''Test FilesystemConsumer'''

    #pylint: disable-msg=R0914

    import random

    stats_dir = os.path.join(base, 'stats')
    events_dir = os.path.join(base, 'events')

    #pylint: disable-msg=W0141
    map(lambda f: os.mkdir(f, 0700), (stats_dir, events_dir))
    #pylint: enable-msg=W0141

    consumer = FilesystemConsumer(stats_dir, events_dir)

    def create_stat():
        '''Create a random stat value'''

        return MonitorDataItem('id%d' % random.randint(1, 100),
                               'type%d' % random.randint(1, 100),
                               ['v', 'a', 'lue', random.randint(1, 100)],
                               long(time.time() * 1000))

    class Severity(object): #pylint: disable-msg=R0903
        '''A simple event severity representation'''

        def __init__(self, severity):
            self._severity = severity

        def __str__(self):
            return self._severity

    Severity.ERROR = Severity('ERROR')
    Severity.CRITICAL = Severity('CRITICAL')

    def create_event():
        '''Create a random event value'''

        return EventItem('id%d' % random.randint(1, 100),
                         'type%d' % random.randint(1, 100),
                         'value%d' % random.randint(1, 100),
                         'log%d' % random.randint(1, 100),
                         random.choice((Severity.ERROR, Severity.CRITICAL, )),
                         long(time.time() * 1000))

    stats = set(create_stat() for _ in xrange(100))
    events = set(create_event() for _ in xrange(100))

    consumer.publish_list(stats.union(events))

    def transform_dict(data):
        '''Transform values of a dict into a tuple, sorted on key'''

        # OVM doesn't ship Python2.5, so we need to expand the loop
        # return tuple(value if not isinstance(value, list) else tuple(value)
        #              for (_, value) in sorted(data.iteritems()))

        result = []

        for (_, value) in sorted(data.iteritems()):
            if not isinstance(value, list):
                result.append(value)
            else:
                result.append(tuple(value))

        return tuple(result)


    stats2 = set(transform_dict(dump_stat(stat_)) for stat_ in stats)
    events2 = set(transform_dict(json.loads(serialize_event(event)))
                  for event in events)

    stats_files = [file_ for file_ in os.listdir(stats_dir)
                   if os.path.isfile(os.path.join(stats_dir, file_))]
    assert len(stats_files) == 1, 'Not one stats file created'

    #pylint: disable-msg=C0103
    fd = open(os.path.join(stats_dir, stats_files[0]), 'r')
    #pylint: enable-msg=C0103
    try:
        stats3 = json.load(fd)
    finally:
        fd.close()

    assert stats3['serialization_version'] == SERIALIZATION_VERSION, \
        'Stats \'serialization_version\' attribute didn\'t survive ' \
        'serialization'
    assert set(transform_dict(stat_) for stat_ in stats3['stats']) == stats2, \
        'Stats didn\'t survive publishing'

    assert len(os.listdir(os.path.join(stats_dir, TEMP_NAME))) == 0, \
        'Stats temp folder is not empty'

    event_files = [file_ for file_ in os.listdir(events_dir)
                   if os.path.isfile(os.path.join(events_dir, file_))]
    assert len(event_files) == len(events), \
        'Invalid number of event files created'

    events3 = set()
    for event_file in event_files:
        #pylint: disable-msg=C0103
        fd = open(os.path.join(events_dir, event_file), 'r')
        #pylint: enable-msg=C0103
        try:
            events3.add(transform_dict(json.load(fd)))
        finally:
            fd.close()

    assert events3 == events2, 'Events didn\'t survive publishing'
    assert len(os.listdir(os.path.join(events_dir, TEMP_NAME))) == 0, \
        'Events temp folder is not empty'


def check_folder(folder):
    '''Assert given folder exists and is writable by the current user

    @param folder: Folder to test
    @type folder: `str`
    '''

    if not os.path.isdir(folder):
        raise RuntimeError('Storage folder %s doesn\'t exist' % folder)

    stat_data = os.stat(folder)

    if (stat_data[stat.ST_MODE] & stat.S_IWRITE) == 0:
        raise RuntimeError('Storage folder %s is not writable' % folder)

@tempdir_test
def test_check_folder_is_dir(base):
    '''Assert check_folder only accepts folders as argument'''

    file_ = os.path.join(base, 'somefile')

    fd = open(file_, 'w') #pylint: disable-msg=C0103
    try:
        fd.write('.')
    finally:
        fd.close()

    try:
        check_folder(file_)
    except RuntimeError:
        return

    assert False, 'check_folder accepts files as argument'

@tempdir_test
def test_check_folder_writable(base):
    '''Assert check_folder only accepts writable folders as argument'''

    dir_ = os.path.join(base, 'somedir')
    os.mkdir(dir_, 0500)

    try:
        check_folder(dir_)
    except RuntimeError:
        return

    assert False, 'check_folder accepts non-writable folders as argument'

@tempdir_test
def test_check_folder(base):
    '''Test valid check_folder behaviour'''

    dir_ = os.path.join(base, 'somedir')
    os.mkdir(dir_, 0700)

    check_folder(dir_)


def create_temp(base):
    '''Create a temp folder in the given base folder

    @param base: Parent folder of temp folder to create
    @type base: `str`
    '''

    tmp_path = os.path.join(base, TEMP_NAME)

    if os.path.isdir(tmp_path):
        check_folder(tmp_path)
    elif os.path.exists(tmp_path):
        raise RuntimeError('Path %s is not a folder' % tmp_path)
    else:
        os.mkdir(tmp_path, 0700)

    return tmp_path

@tempdir_test
def test_create_temp_is_file(base):
    '''Assert create_temp fails if tmp exists, but is not a folder'''

    file_ = os.path.join(base, TEMP_NAME)

    fd = open(file_, 'w') #pylint: disable-msg=C0103
    try:
        fd.write('.')
    finally:
        fd.close()

    try:
        create_temp(base)
    except RuntimeError:
        return

    assert False, 'create_temp doesn\'t fail if TEMP_NAME is a file'

@tempdir_test
def test_create_temp_is_dir(base):
    '''Assert create_temp succeeds if tmp exists and is a writable folder'''

    dir_ = os.path.join(base, TEMP_NAME)
    os.mkdir(dir_, 0700)

    dir2 = create_temp(base)

    assert dir_ == dir2, 'create_temp returned unexpected value'

@tempdir_test
def test_create_temp_is_non_writable_dir(base): #pylint: disable-msg=C0103
    '''Assert create_temp fails if tmp exists but is a non-writable folder'''

    dir_ = os.path.join(base, TEMP_NAME)
    os.mkdir(dir_, 0500)

    try:
        create_temp(base)
    except RuntimeError:
        return

    assert False, \
        'create_temp doesn\'t fail if tmp dir exists but is not writable'

@tempdir_test
def test_create_temp(base):
    '''Test valid create_temp behaviour'''

    dir_ = os.path.join(base, TEMP_NAME)

    dir2 = create_temp(base)

    assert dir_ == dir2, 'create_temp returned unexpected value'
    assert os.path.exists(dir_), 'create_temp didn\'t create expected tmp dir'
    assert os.path.isdir(dir_), 'create_temp didn\'t create expected tmp dir'
    assert os.stat(dir_)[stat.ST_MODE] & stat.S_IWRITE != 0, \
        'create_temp created a non-writable tmp dir'


def write_temp(data, base, filename=None):
    '''Write L{data} to a temporary file in L{base}

    The path of the created temporary file is returned.

    @param data: Data to write into the temporary file
    @type data: `str` or `unicode`
    @param base: Base folder in which to create the temporary file
    @type base: `str`

    @return: Temporary file path
    @rtype: `str`
    '''

    if isinstance(data, unicode):
        data = data.encode('utf-8')

    if filename:
        name = os.path.join(base, filename)
        handle = open(name, 'w')
    else:
        fd, name = tempfile.mkstemp(dir=base) #pylint: disable-msg=C0103
        handle = os.fdopen(fd, 'w')
    try:
        handle.write(data)
    finally:
        handle.close()

    return name

@tempdir_test
def test_write_temp(base):
    '''Test valid write_temp behaviour'''

    import random
    import string #pylint: disable-msg=W0402

    msg_ = ''.join(random.choice(string.printable)
                   for _ in xrange(random.randint(512, 1024)))

    file_ = write_temp(msg_, base)

    assert os.path.isfile(file_), 'write_temp didn\'t create a file'

    fd = open(file_, 'r') #pylint: disable-msg=C0103
    try:
        msg2 = fd.read()
    finally:
        fd.close()

    assert msg2 == msg_, 'write_file didn\'t write the given message'

@tempdir_test
def test_write_unicode(base):
    '''Test valid UTF8 write_temp behaviour'''

    msg_ = u''.join(unichr(i) for i in xrange(10000, 10010))

    file_ = write_temp(msg_, base)

    assert os.path.isfile(file_), 'write_temp didn\'t create a file'

    fd = open(file_, 'r') # pylint: disable-msg=C0103
    try:
        msg2 = fd.read()
    finally:
        fd.close()

    msg3 = msg2.decode('utf-8')
    assert type(msg3) is unicode, 'Non-unicode returned by string.decode'

    assert msg3 == msg_, 'write_file didn\'t write the given unicode message'


def relink(original, target_base):
    '''Relink a given file into another folder

    @param original: Path to original file to relink
    @type original: `str`
    @param target_base: Target folder path
    @type target_base: `str`

    @return: New file path
    @rtype: `str`
    '''

    _, name = os.path.split(original)
    target = os.path.join(target_base, name)

    os.rename(original, target)

    return target

@tempdir_test
def test_relink(base):
    '''Test valid relink behaviour'''

    import random
    import string #pylint: disable-msg=W0402

    msg_ = ''.join(random.choice(string.printable)
                   for _ in xrange(random.randint(512, 1024)))

    tmp = os.path.join(base, TEMP_NAME)
    os.mkdir(tmp, 0700)

    fd, file_ = tempfile.mkstemp(dir=tmp) #pylint: disable-msg=C0103
    handle = os.fdopen(fd, 'w')

    try:
        handle.write(msg_)
    finally:
        handle.close()

    relinked = relink(file_, base)

    assert os.path.isfile(relinked), 'relink returned a non-file path'

    fd2 = open(relinked, 'r')
    try:
        msg2 = fd2.read()
    finally:
        fd2.close()

    assert msg2 == msg_, 'relink changed file content'
    assert os.path.dirname(relinked) == base, \
        'relink didn\'t put the file in base'
    assert not os.path.exists(file_), 'relink didn\'t remove the original file'


def serialize_event(event):
    '''Serialize an event using JSON serialization

    @param event: Event to serialize
    @type event: L{EventItem}

    @return: JSON representation of the event
    @rtype: `unicode`
    '''

    return json.dumps({
        'serialization_version': SERIALIZATION_VERSION,
        'id': event.id,
        'type': event.type,
        'value': event.value,
        'log': event.log,
        'severity': str(event.severity),
        'timestamp': event.timestamp,
    })

def test_serialize_event():
    '''Test event serialization'''

    class Severity(object): #pylint: disable-msg=R0903
        '''A simple event severity representation'''

        def __init__(self, severity):
            self._severity = severity

        def __str__(self):
            return self._severity

    Severity.CRITICAL = Severity('CRITICAL')

    timestamp = long(time.time() * 1000)

    event = EventItem('id', 'type', 'value', 'log', Severity.CRITICAL, timestamp)

    serialized_event = serialize_event(event)

    event2 = json.loads(serialized_event)

    for key in ('id', 'type', 'value', 'log', ):
        assert event2[key] == getattr(event, key), \
            'Event attribute \'%s\' didn\'t survive serialization' % key

    assert event2['severity'] == 'CRITICAL', \
        'Event attribute \'severity\' didn\'t survive serialization'
    assert event2['timestamp'] == timestamp, \
        'Event attribute \'timestamp\' didn\'t survive serialization'
    assert event2['serialization_version'] == SERIALIZATION_VERSION, \
        'Event attribute \'serialization_version\' didn\'t survive ' \
        'serialization'


def dump_stat(stat_):
    '''Dump a statistic value into a dictionary representation

    @param stat_: Statistic value to dump
    @type stat_: L{MonitorDataItem}

    @return: Dictionary containing the statistic value
    @rtype: `dict`
    '''

    return {
        'id': stat_.id,
        'type': stat_.type,
        'value': stat_.value,
        'timestamp': stat_.timestamp,
    }

def test_dump_stat():
    '''Test stat dumping'''

    timestamp = long(time.time() * 1000)

    stat_ = MonitorDataItem('id', 'type', ['v', 'a', 'lue'], timestamp)

    stat2 = dump_stat(stat_)

    for key in ('id', 'type', 'value', ):
        assert stat2[key] == getattr(stat_, key), \
            'Stat attribute \'%s\' didn\'t survive dumping' % key

    assert stat2['timestamp'] == timestamp, \
        'Stat attribute \'timestamp\' didn\'t survive dumping'


def serialize_stats(stats):
    '''Serialize statistics using JSON serialization

    @param stats: Stats to serialize
    @type stats: `iterable`

    @return: JSON representation of the statistics
    @rtype: `unicode`
    '''

    return json.dumps({
        'serialization_version': SERIALIZATION_VERSION,
        'stats': tuple(stats),
    })

def test_serialize_stats():
    '''Test stat serialization'''

    import random

    def create_stat():
        '''Create a dumped stat dictionary'''

        return {
            'id': 'id%d' % random.randint(1, 100),
            'type': 'type%d' % random.randint(1, 100),
            'value': 'value%d' % random.randint(1, 100),
            'timestamp': long(time.time() * 1000),
        }

    stats = [create_stat() for _ in xrange(10)]

    serialized_stats = serialize_stats(stats)

    stats2 = json.loads(serialized_stats)

    assert stats2['serialization_version'] == SERIALIZATION_VERSION, \
        'Stats attribute \'serialization_version\' didn\'t survive ' \
        'serialization'
    assert stats2['stats'] == stats, \
        'Stats attribute \'stats\' didn\'t survive serialization'
