# Copyright (C) 2011 Oracle. All rights reserved.
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

'''Monitor data consumer sending all items to an XMPP node'''

import uuid
import math
import types
import logging
import functools

try:
    import json
except ImportError:
    import simplejson as json

from twisted.internet import defer, reactor
from twisted.python import log
from twisted.words.xish import domish
from twisted.words.protocols.jabber import jid as jid_
from twisted.words.protocols.jabber.sasl import SASLAuthError
from twisted.words.protocols.jabber.xmlstream import toResponse

from wokkel.client import XMPPClient
from wokkel.xmppim import AvailablePresence
from wokkel.subprotocols import XMPPHandler
from wokkel.generic import VersionHandler, IQ_GET
from wokkel.disco import DiscoHandler, IDisco, DiscoFeature

from zope.interface import implements

import monitor
from monitor.consumers.basic import BufferedConsumer
from monitor.service import MonitorDataItem, EventItem

#pylint: disable-msg=C0103

msg = functools.partial(log.msg, system='monitor.consumer.xmpp')

NS_PING = 'urn:xmpp:ping'
PING = IQ_GET + '/ping[@xmlns="' + NS_PING + '"]'
class PingHandler(XMPPHandler):
    '''XMPP subprotocol handler for XMPP Ping

    This protocol is described in
    U{XEP-199<http://www.xmpp.org/extensions/xep-0199.html>}.
    '''
    implements(IDisco)

    def connectionInitialized(self):
        self.xmlstream.addObserver(PING, self.onPing)

    # XMPP Ping handler
    def onPing(self, iq):
        log.msg('Got XMPP Ping', system='monitor.consumer.xmpp',
                level=logging.DEBUG)
        response = toResponse(iq, 'result')
        self.send(response)
        iq.handled = True

    # IDisco methods
    def getDiscoInfo(self, requestor, target, node):
        info = set()

        if not node:
            info.add(DiscoFeature(NS_PING))

        return defer.succeed(info)

    def getDiscoItems(self, requestor, target, node):
        return defer.succeed([])


class XMPPConsumerProtocol(XMPPHandler):
    '''XMPP consumer Twisted/Wokkel protocol'''
    NS_BASE = 'http://cloud.sun.com/ns/nephos/monitor'
    NS = '%s/1' % NS_BASE

    implements(IDisco)

    def __init__(self, from_, destination):
        XMPPHandler.__init__(self)
        self.from_ = from_
        self.destination = destination
        self.thread = str(uuid.uuid4())

    def connectionMade(self):
        '''A connection with the server has been made'''
        msg('XMPP consumer %s connected' % self.from_)
        XMPPHandler.connectionMade(self)
        self.send(AvailablePresence()) #pylint: disable-msg=E1101

    def connectionLost(self, reason):
        '''Connection with the server was lost'''
        msg('XMPP consumer %s lost connection' % self.from_)
        XMPPHandler.connectionLost(self, reason)

    def sendItems(self, items):
        '''Send a couple of items to C{destination}
        
        @param items: Iterable of monitor items
        @type items: iterable
        '''
        items = tuple(items)

        message = domish.Element((None, 'message'))
        message['from'] = self.from_
        message['to'] = self.destination
        message['type'] = 'normal'

        message.addElement('thread', content=self.thread)
        message.addElement('body', content='%d monitor item(s)' % len(items))

        item_container = message.addElement((self.NS, 'items', ))

        for item in items:
            if isinstance(item, MonitorDataItem):
                self._monitorDataItemToElement(item_container, item)
            elif isinstance(item, EventItem):
                self._eventItemToElement(item_container, item)
            else:
                raise TypeError('Unknown item type %s' % type(item))

        self.send(message) #pylint: disable-msg=E1101

    @classmethod
    def _monitorDataItemToElement(cls, parent, item):
        '''Attach a C{MonitorDataItem} to a domish element'''
        elem = parent.addElement((None, 'monitor_data', ))
        elem.addElement('id', content=item.id)
        elem.addElement('type', content=item.type)

        value = elem.addElement((None, 'value'))
        value['encoding'] = 'json'
        value_content = json.dumps(item.value)
        value.addContent(value_content)

        elem.addElement('timestamp', content=item.timestamp)

    @classmethod
    def _eventItemToElement(cls, parent, item):
        '''Attach a C{EventItem} to a domish element'''
        elem = parent.addElement((None, 'event', ))

        elem.addElement('id', content=item.id)
        elem.addElement('type', content=item.type)
        if item.value is not None:
            value = elem.addElement((None, 'value'))
            value['encoding'] = 'json'
            value_content = json.dumps(item.value)
            value.addContent(value_content)
        elem.addElement('log', content=item.log)
        elem.addElement('severity', content=str(item.severity))
        elem.addElement('timestamp', content=item.timestamp)

    # IDisco methods
    def getDiscoInfo(self, requestor, target, node):
        info = set()

        if not node:
            info.add(DiscoFeature(self.NS))

        return defer.succeed(info)

    def getDiscoItems(self, requestor, target, node):
        return defer.succeed([])


class ReconnectingXMPPClient(XMPPClient):
    '''XMPPClient modified to reconnect after authentication failure'''
    RECONNECT_DELAY = 5

    _connection = None

    def initializationFailed(self, reason):
        '''Called when stream initialization has failed.

        Stop the service (thereby disconnecting the current stream) and
        raise the exception, unless the disconnect reason was a SASL
        authentication error, in which case a reconnection attempt will be
        scheduled.
        '''
        if getattr(reason, 'type', None) is not SASLAuthError:
            return XMPPClient.initializationFailed(self, reason)

        log.err(reason,
                'XMPP SASL authentication failed, scheduling reconnect')

        # Cancel current connection and stop factory
        self.factory.stopTrying()
        if self._connection:
            try:
                self._connection.disconnect()
            except Exception, exc:
                log.err(exc, 'Failed to disconnect from XMPP server')

        # Schedule reconnect
        def reconnect():
            '''Reconnect to XMPP server'''
            log.msg('Reconnecting XMPP connection after SASL ' \
                    'authentication failure')
            self.factory.resetDelay()
            self._connection = self._getConnection()

        reactor.callLater(self.RECONNECT_DELAY, reconnect)


class XMPPConsumer(BufferedConsumer):
    '''Monitor data consumer sending data items to an XMPP network/node'''
    #pylint: disable-msg=R0913
    def __init__(self, jid, password, destination, host=None, port=5222,
                 buffer_timeout=5, buffer_events=False, debug=False):
        '''Initialize a new C{XMPPConsumer}

        @param jid: JID to use for authentication
        @type jid: str
        @param password: Password to use for authentication
        @type password: str
        @param destination: Destination JID for monitoring items
        @type destination: str
        @param host: Host to connect to (use DNS SRV info if C{None})
        @type host: str
        @param port: Port to connect to
        @type port: number
        @param buffer_timeout: Buffer flush timeout (in seconds)
        @type buffer_timeout: number
        @param buffer_events: Buffer events
        @type buffer_events: bool
        @param debug: Produce debugging packet output
        @type debug: bool
        '''
        msg('Creating XMPP consumer for JID %s' % jid)
        self.jid = jid
        self.destination = destination

        self.client = ReconnectingXMPPClient(jid_.JID(jid), password, host,
                                             port)

        # Limit maximum backoff time between reconnection attempts
        self.client.factory.maxDelay = math.sqrt(2) * math.sqrt(3) * \
                                        math.sqrt(5)

        if debug:
            self.client.logTraffic = True

        # Version handler - XEP92
        protocol = VersionHandler('Nephos Monitor',
                              '.'.join(str(i) for i in monitor.__version__))
        protocol.setHandlerParent(self.client)

        # Disco handler - XEP30
        protocol = DiscoHandler()
        protocol.setHandlerParent(self.client)

        # Ping handler - XEP199
        protocol = PingHandler()
        protocol.setHandlerParent(self.client)

        self.proto = XMPPConsumerProtocol(jid, destination)
        self.proto.setHandlerParent(self.client) #pylint: disable-msg=E1101

        BufferedConsumer.__init__(self, buffer_timeout, buffer_events)

    def startService(self):
        '''Start the service'''
        self.client.startService()
        BufferedConsumer.startService(self)

    def stopService(self):
        '''Stop the service'''
        self.flush()
        BufferedConsumer.stopService(self)
        self.client.stopService()

    def publish_list(self, data):
        '''Publish a list of monitor items

        This is required by the abstract L{BufferedConsumer} class.

        @param data: Iterable of monitor data items
        @type data: iterable
        '''
        self.proto.sendItems(data)
