# Copyright (C) 2011 Oracle. All rights reserved.
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

from monitor.producer import loop
import os.path
import subprocess

log = lambda *_: None

FALLBACK_MACHINE_ID = 'localhost'

procStatParameters = \
    (('pid', int), 
     ('comm', str),
     ('state', str),
     ('ppid', int),
     ('pgrp', int),
     ('session', int),
     ('tty_nr', int),
     ('tpgid', int),
     ('flags', int), 
     ('minflt', int),
     ('cminflt', int),
     ('majflt', int),
     ('cmajflt', int),
     ('utime', int),
     ('stime', int),
     ('cutime', int),
     ('cstime', int),
     ('priority', int),
     ('nice', int),
     ('num_threads', int),
     ('itrealvalue', int),
     ('starttime', int),
     ('vsize', int),
     ('rss', int),
     ('rlim', int),
     ('startcode', int),
     ('endcode', int),
     ('startstack', int),
     ('kstkesp', int),
     ('kstkeip', int),
     ('signal', int),
     ('blocked', int),
     ('sigignore', int),
     ('sigcatch', int),
     ('wchan', int),
     ('nswap', int),
     ('cnswap', int),
     ('exit_signal', int),
     ('processor', int),
     ('rt_priority', int),
     ('policy', int),
     ('delayacct_blkio_ticks', int))
memParameters = ('MemFree', 'MemTotal', 'SwapFree', 'SwapTotal', 'Buffers', 'Cached')
strDigits = map(str, xrange(10))

def stats(pidFilename):
    if not os.path.exists(pidFilename):
        return False
    f = open(pidFilename, 'r')

    try:
        try:
            pid = int(f.readline().strip())
        except ValueError:
            log("Cannot parse pid file %s" % pidFilename)
            return False
    finally:
        f.close()

    statsFilename = os.path.join('/proc', str(pid), 'stat')
    if not os.path.exists(statsFilename):
        log("No stats found in %s for pid file %s" % (statsFilename, pidFilename))
        return False
    f = open(statsFilename, 'r')
    try:
        stats = f.readline().split(' ')
    finally:
        f.close()
    return dict(map(lambda i: (procStatParameters[i][0], procStatParameters[i][1](stats[i])), xrange(len(stats))))
    

def meminfo():
    f = open('/proc/meminfo', 'r')
    try:
        return dict([ci.strip() for ci in i.split(':')] for i in f.readlines())
    finally:
        f.close()

_physCpusOnly = lambda b:(b[0]!='cpu' and b[0].startswith('cpu'))
_stripAndSplit = lambda a:a.replace('  ',' ').split(' ')
def cpuStats():
    cpustat = {}
    f = open('/proc/stat', 'r')
    try:
        xx = filter(_physCpusOnly, map(_stripAndSplit, f.readlines()))
        for cpu,usr,nice,sys,idle,_,_,_,_,_ in xx:
            cpustat[cpu] = {'usr': usr, 'nice': nice, 'sys': sys, 'idle': idle}
    finally:
        f.close()
    return cpustat

def uptime():
    f = open('/proc/uptime', 'r')
    try:
        return float(f.readline().split(' ')[0])
    finally:
        f.close()

def load():
    f = open('/proc/loadavg', 'r')
    try:
        return float(f.readline().split(' ')[0])
    finally:
        f.close()

_statLineToList = lambda line: filter(lambda x: x!='', line.strip().split(' '))
def diskstats():
    readret = {}
    writeret = {}
    f = open('/proc/diskstats', 'r')
    try:
        data = map(_statLineToList, f.readlines())
        data = filter(lambda x: x[2].startswith('dm-'), data)
        data = map_dm_to_uuid(data)
        for stats in data:
            readret[stats[2]] = int(stats[5]) << 9
            writeret[stats[2]] = int(stats[9]) << 9
        return readret,writeret
    finally:
        f.close()

def map_dm_to_uuid(data):
    for stats in data:
        cmd = ["dmsetup", "info", "--columns", "--noheadings", "--options=UUID", "/dev/%s" % stats[2]]
        dms_p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (p_out, p_err) = dms_p.communicate()

        if dms_p.returncode != 0:
            log("Unable to map multipath device cmd: (%s): returncode: (%s), error: (%s)" % (cmd, dms_p.returncode, p_err))
            data.remove(stats)
            continue

        if p_out.startswith('part'):
            data.remove(stats)
            continue

        p_out = p_out.rstrip("\n").replace("mpath-", "")
        stats[2] = p_out

    return (data)



_physicalNics = lambda stats: (not 'lo' in stats[0] and not stats[0].startswith('vif'))
_vmNics = lambda stats: stats[0].startswith('vif')
_nicStats = lambda stats: stats[0:3] + stats[9:11]
_replaceColonWithSpace = lambda line: line.replace(':', ' ')
def netinfo():
    f = open('/proc/net/dev', 'r')
    try:
        data = f.readlines()[2:]
        netstats = map(_nicStats, filter(_physicalNics, map(_statLineToList, map(_replaceColonWithSpace, data) ) ) )
        srv_result = { 'serverPortSent': { }, 'serverPortReceive': { } }
        for nic, stats in ((n[0], n[1:]) for n in netstats):
            rec_bytes, rec_packets, sen_bytes, sen_packets = (int(s) for s in stats)
            srv_result['serverPortSent'][nic] = sen_bytes
            srv_result['serverPortReceive'][nic] = rec_bytes

        netstats = map(_nicStats, filter(_vmNics, map(_statLineToList, map(_replaceColonWithSpace, data) ) ) )
        vm_result = { 'vmNetworkWrite': { }, 'vmNetworkReceive': { } }
        for nic, stats in ((n[0], n[1:]) for n in netstats):
            rec_bytes, rec_packets, sen_bytes, sen_packets = (int(s) for s in stats)
            vm_result['vmNetworkWrite'][nic] = sen_bytes
            vm_result['vmNetworkReceive'][nic] = rec_bytes
        return srv_result,vm_result
    finally:
        f.close()


def produce():
    '''Produce monitoring data'''
    try:
        from monitor.producer import nephos_config
    except ImportError:
        machine_id = FALLBACK_MACHINE_ID
    else:
        machine_id = nephos_config.machineid

    try:
        # memory monitors
        values = meminfo()
        yield machine_id, 'serverFreeMem', values['MemFree']
        yield machine_id, 'serverFreeSwap', values['SwapFree']
        # load monitor
        yield machine_id, 'serverCpuStats', cpuStats()
        yield machine_id, 'serverCpuLoad', load()
        # uptime
        yield machine_id, 'serverUptime', uptime()
        # disk monitor
        r,w = diskstats()
        yield machine_id, 'serverDiskRead', r
        yield machine_id, 'serverDiskWrite', w
        # network info monitor
        srv,vm = netinfo()
        yield machine_id, 'serverPortSent', srv['serverPortSent']
        yield machine_id, 'serverPortReceive', srv['serverPortReceive']
        yield machine_id, 'vmNetworkWrite', vm['vmNetworkWrite']
        yield machine_id, 'vmNetworkReceive', vm['vmNetworkReceive']
    except Exception, e:
        pass


if __name__ == '__main__':
    loop(produce)
