# Copyright (C) 2011 Oracle. All rights reserved.
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

'''
Monitoring plugin using the Xen Python bindings to retrieve hypervisor and
domain information
'''

import time
import itertools
import threading
import os

from monitor.producer import loop, InvalidPlatformException

try:
    import xen.xend.XendDomain
    import xen.xend.XendClient
    from xen.xend.xenstore.xstransact import xstransact
except ImportError, e:
    raise InvalidPlatformException('Unable to import Xen modules')


# The ID of the local host. Change this to some other mechanism to retrieve this
# ID if available.
HOST_ID = 'localhost'

# Map Xend state IDs to human-readable states
DOM_STATE_MAP = {
    xen.xend.XendDomain.DOM_STATE_CRASHED: 'CRASHED',
    xen.xend.XendDomain.DOM_STATE_HALTED: 'HALTED',
    xen.xend.XendDomain.DOM_STATE_PAUSED: 'PAUSED',
    xen.xend.XendDomain.DOM_STATE_RUNNING: 'RUNNING',
    xen.xend.XendDomain.DOM_STATE_SHUTDOWN: 'SHUTDOWN',
    xen.xend.XendDomain.DOM_STATE_SUSPENDED: 'SUSPENDED',
    xen.xend.XendDomain.DOM_STATE_UNKNOWN: 'UNKNOWN',
}

# Loop delay for the VCPU usage counter thread
CPU_USAGE_COUNTER_DELAY = 10


class Break(Exception):
    '''Exception to be thrown to break a forever loop'''

    pass

def forever(fun):
    '''Wrap a callable so it runs forever

    Exceptions (except KeyboardInterrupt and Break) are caught and ignored.
    The function is executed in a continuous while-loop.

    The loop can be interrupted by raising Break from the wrapped function.
    '''

    def wrapped(*args, **kwargs):
        while True:
            try:
                fun(*args, **kwargs)
            except Break:
                break
            except KeyboardInterrupt:
                raise
            except Exception, ex:
                # Oops
                pass

    return wrapped


class CPUUsageCounter(threading.Thread):
    '''
    A background thread keeping track of (V)CPU usage percentages of Xen
    domains
    '''

    def __init__(self):
        threading.Thread.__init__(self)
        # This needs to be a daemon thread
        self.setDaemon(True)

        # Timestamp of the previous iteration
        self._previous_run = -1
        # Previous counters
        self._previous_usages = {}
        # Current usage percentages
        self._values = {}

    def __getitem__(self, key):
        return self._values[key]

    @forever
    def run(self):
        # This is based on get_cpu_pct from xentop.c, although it might not be
        # completely correct...

        # Get current timestamp
        timestamp = time.time()
        # Calculate delta to previous round
        time_delta = timestamp - self._previous_run

        try:
            # Retrieve, for every Xen domain, the sum of all cpu_time stats for
            # every VCPU they use
            cpu_usage = self._get_cpu_usage()
        except Exception, ex:
            # Oops
            pass
        else:
            # Delete non-existing domains from the storage dict
            map(self._values.pop,
                set(self._values.iterkeys()).difference(cpu_usage.iterkeys()))

            # Unless this is the first run
            if self._previous_run >= 0:
                # For every domain
                for key in cpu_usage:
                    cpu_time = cpu_usage[key][0]
                    cpu_count = cpu_usage[key][1]
                    # Retrieve the previous CPU usage
                    old_cpu_time = self._previous_usages.get(key, (0.0, 1))[0]
                    # Calculate usage delta
                    usage_delta = cpu_time - old_cpu_time
                    # Calculate usage change per time unit percentage
                    load = usage_delta * 100 / time_delta
                    ceiling = float(cpu_count * 100)
                    if load > ceiling:
                        load = ceiling
                    util = float(load / (cpu_count * 100))
                    self._values[key] = (load, util)

            # Store for next cycle
            self._previous_usages = cpu_usage
            self._previous_run = timestamp

        # Sleep a bit
        time.sleep(CPU_USAGE_COUNTER_DELAY)

    @staticmethod
    def _get_cpu_usage():
        '''Calculate aggregated VCPU usage for all domains'''

        # Helper
        filter_ = lambda key: lambda values: \
            (value[1:] for value in values if value[0] == key)

        # Filter cpu_time fields
        cpu_time_filter = filter_('cpu_time')
        # Filter vcpu_count fields
        vcpu_count_filter = filter_('vcpu_count')
        # Map cpu_time filter over all fields
        cpu_times = lambda info: (cpu_time_filter(c) for c in info)
        # Filter vcpu fields
        vcpus = filter_('vcpu')

        # Flatten an iterable
        # Somewhat like (flatmap id):
        # [[1], [2, 3], [4]] becomes [1, 2, 3, 4]
        flatten = lambda v: itertools.chain(*v)

        # Aggregate (sum) all (V)CPU usage times
        aggregate = lambda info: \
            float(sum(flatten(flatten(cpu_times(vcpus(info))))))

        def vcpu_count(info):
            for c in vcpu_count_filter(info):
                if c:
                    return int(c[0])
            else:
                return 1

        domains = xen.xend.XendClient.server.xend.domains
        getVCPUInfo = xen.xend.XendClient.server.xend.domain.getVCPUInfo

        # Set up result dict
        return dict((domain, (aggregate(getVCPUInfo(domain)), vcpu_count(getVCPUInfo(domain))))
            for domain in domains(False))


def vmdiskstats(domain_id):
    vbds_reads = {}
    vbds_writes = {}
    vmdevs = '/sys/bus/xen-backend/devices'
    try:
        for entry in os.listdir(vmdevs):
            if not entry.startswith('vbd-%s' % domain_id):
                continue

            f = open(os.path.join(vmdevs, entry, 'nodename'))
            try:
                nodename = f.read().strip()
            finally:
                f.close()

            try:
                backend = xstransact.Read('/local/domain/0/%s/params' % nodename).strip()
            except Exception:
                continue

            if not backend:
                continue

            # remove leading 'aio:', 'ioemu:', etc.
            backend = backend.split(':', 1)[-1]

            f = open(os.path.join(vmdevs, entry, 'statistics/rd_sect'))
            try:
                reads = f.readline().strip()
            finally:
                f.close()

            f = open(os.path.join(vmdevs, entry, 'statistics/wr_sect'))
            try:
                writes = f.readline().strip()
            finally:
                f.close()

            vbds_reads[backend] = reads
            vbds_writes[backend] = writes
    except Exception:
       pass

    return vbds_reads, vbds_writes


STATNAMES = { 'cpu_time': 'vmCpuTime',
              'domid': 'vmDomid',
              'maxmem': 'vmMaxMemory',
              'memory': 'vmCurrentMemory',
              'online_vcpus': 'vmOnlineVcpus',
              'state': 'vmState',
              'status': 'vmStatus',
              'vcpus': 'vmVcpus',
              'uuid': 'vmUUID',
              'domain_count': 'vmDomainCount',
              'total_memory': 'vmDom0TotalMemory',
              'free_memory': 'vmDom0FreeMemory',
              'cpu_load': 'vmCpuLoad',
              'cpu_util': 'vmCpuUtil' }

def produce_domain_info(domain, cpu_counter):
    '''Produce statistics of one domain'''

    # Turn data into a dict for ease of use
    domain_info = dict(domain[1:])

    # ID of the domain
    id_ = '%s/%s' % (HOST_ID, domain_info['name'])

    # Yield all known stat values
    for item in 'cpu_time', 'domid', 'maxmem', 'memory', \
        'online_vcpus', 'state', 'vcpus', 'uuid':
        if item in domain_info:
            yield id_, STATNAMES[item], domain_info[item]

    # Yield vdb stats
    if domain_info['name'] != 'Domain-0':
        vbds_reads, vbds_writes = vmdiskstats(domain_info['domid'])
        if vbds_reads:
            yield id_, 'vmDiskRead', vbds_reads
        if vbds_writes:
            yield id_, 'vmDiskWrite', vbds_writes


    # Yield human-readable domain status, if available
    if 'status' in domain_info:
        yield id_, STATNAMES['status'], \
            DOM_STATE_MAP.get(domain_info['status'], 'UNKNOWN')

    # Yield (V)CPU usage, if available
    try:
        cpu_load, cpu_util = cpu_counter[domain_info['name']]
        yield id_, STATNAMES['cpu_load'], cpu_load
        yield id_, STATNAMES['cpu_util'], cpu_util
    except KeyError:
        pass



def produce(cpu_counter):
    '''Produce Xen statistics'''

    try:
        # Retrieve domain list
        domains = xen.xend.XendClient.server.xend.domains(1)
        # Retrieve and parse node info
        node_info = dict(xen.xend.XendClient.server.xend.node.info())
    
        # Yield number of domains running on this node
        yield HOST_ID, STATNAMES['domain_count'], len(domains)
    
        # Yield the total memory available on the node
        if 'total_memory' in node_info:
            yield HOST_ID, STATNAMES['total_memory'], \
                int(node_info['total_memory'])
        # Yield the free memory available on the node
        if 'free_memory' in node_info:
            yield HOST_ID, STATNAMES['free_memory'], \
                int(node_info['free_memory'])
    
        # For every domain
        for domain in domains:
            # Yield its stats
            for item in produce_domain_info(domain, cpu_counter):
                yield item
    except Exception, e:
        pass


if __name__ == '__main__':
    # Launch a CPU usage counter thread
    cputime_thread = CPUUsageCounter()
    cputime_thread.start()

    # Start stat production loop
    loop(lambda: produce(cputime_thread))
