# Copyright (C) 2011 Oracle. All rights reserved.
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

'''Library methods for monitor data producers'''

import os
import sys
import time
import types
import optparse

try:
    import json
except ImportError:
    import simplejson as json

import yaml

from monitor import __version__ #pylint: disable-msg=W0611

# Python < 2.5 compatibility stuff
try:
    from struct import Struct
except ImportError:
    import struct

    class Struct:
        def __init__(self, format):
            self.format = format

        def pack(self, *args):
            return struct.pack(self.format, *args)


def _create_enum(name, *names): #pylint: disable-msg=W0621
    '''Create an enumeration-style class

    @param name: Name of the class
    @type name: str
    @param names: Names of enumeration items
    @type names: iterable<str>
    '''
    attrs = dict()

    def __init__(self, name): #pylint: disable-msg=W0621
        self._name = name
    attrs['__init__'] = __init__

    def __str__(self):
        return self._name #pylint: disable-msg=W0212
    attrs['__str__'] = __str__

    def __repr__(self):
        return self._name #pylint: disable-msg=W0212
    attrs['__repr__'] = __repr__

    attrs['__slots__'] = ('_name', )
    attrs['__doc__'] = '%s enumeration' % name

    cls = types.ClassType(name, (object, ), attrs)

    for name in names:
        item = cls(name)
        setattr(cls, name, item)

    return cls

#pylint: disable-msg=C0103
levels = 'NOT_SET', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'
SeverityLevel = _create_enum('SeverityLevel', #pylint: disable-msg=W0142
                             *levels)
del _create_enum, levels


class InvalidPlatformException(Exception):
    '''Exception raised when plugin is started on an invalid platform'''

    def __init__(self, *args):
        Exception.__init__(self, *args)

        # Python < 2.5 compatibility
        if not hasattr(self, 'message'):
            if args:
                self.message = args[0]
            else:
                self.message = ''


class Producer(object):
    '''Data producer class handling all IPC'''

    def __init__(self):
        '''Initialize a new helper'''
        #pylint: disable-msg=W0621
        data_fd, interval, debug, name, config_file = self.parse_options()
        self.debug = debug
        self.data_fd = None
        if not debug:
            self.data_fd = os.fdopen(data_fd, 'w')

        self.interval = interval
        self.name = name
        self.config_file = config_file
        self._packer = Struct('<L')#pylint: disable-msg=E0602

    #pylint: disable-msg=R0913
    def produce(self, id_, type_, value, timestamp=None):
        '''Produce one data entity and send it to the monitor application

        @param id_: ID of the entity to which this data relates
        @type id_: str
        @param type_: Type of monitoring information
        @type type_: str
        @param value: Monitored value
        @type value: object
        @param timestamp: Timestamp of value capturing, now() if not set
        @type timestamp: long
        '''
        if timestamp is None:
            timestamp = long(time.time() * 1000)

        data = ('monitor_data_1', id_, type_, value, timestamp)

        if self.debug:
            # Make sure we can serialize the output. If not, the user will see
            # something's going wrong
            try:
                json.dumps(data)
            except:
                sys.stderr.write('\n')
                sys.stderr.write('Error: unable to serialize data item\n')
                sys.stderr.write('\n')
                sys.stderr.write('    ID: %s\n' % id_)
                sys.stderr.write('    Type: %s\n' % type_)
                sys.stderr.write('    Value: %s\n' % value)
                sys.stderr.write('\n')
                raise

            print '[%s] %s: %s = %s' % (timestamp, id_, type_, value)

            return

        serialized = json.dumps(data)
        self.data_fd.write(self._packer.pack(len(serialized)))
        self.data_fd.write(serialized)
        self.data_fd.flush()

    def notify(self, id_, type_, value=None, log='',
               severity=SeverityLevel.NOT_SET, timestamp=None):
        '''Produce one event entity and send it to the monitor application

        @param id_: ID of the entity to which this data relates
        @type id_: str
        @param type_: Type of event information
        @type type_: str
        @param value: Value item of the event, set to None if not required
        @type value: object
        @param log: Logging information related to this event
        @type log: str
        @param severity: Event severity
        @type severity: C{SeverityLevel}
        @param timestamp: Timestamp of value capturing
        @type timestamp: long
        '''
        assert isinstance(severity, SeverityLevel)
        if timestamp is None:
            timestamp = long(time.time() * 1000)

        data = ('event_2', id_, type_, value, log, str(severity), timestamp)

        if self.debug:
            # Make sure we can serialize the output. If not, the user will see
            # something's going wrong
            try:
                json.dumps(data)
            except:
                sys.stderr.write('\n')
                sys.stderr.write('Error: unable to serialize event item\n')
                sys.stderr.write('\n')
                sys.stderr.write('    ID: %s\n' % id_)
                sys.stderr.write('    Type: %s\n' % type_)
                sys.stderr.write('    Value: %s\n' % value)
                sys.stderr.write('    Log: %s\n' % log)
                sys.stderr.write('    Severity: %s\n' % str(severity))
                sys.stderr.write('\n')
                raise

            print '[%s] %s: Event %s (%s): %s' % (timestamp, id_,
                                                  type_,
                                                  str(severity).lower(), log)

            return

        serialized = json.dumps(data)
        self.data_fd.write(self._packer.pack(len(serialized)))
        self.data_fd.write(serialized)
        self.data_fd.flush()

    @staticmethod
    def parse_options():
        '''Parse command line options'''
        parser = optparse.OptionParser()
        parser.add_option('--output-fd', dest='output_fd', type='int',
                          help='file descriptor to write results to',
                          metavar='FD')
        parser.add_option('--interval', dest='interval', type='float',
                          help='producer loop interval', metavar='SEC')
        parser.add_option('-d', '--debug', dest='debug', action='store_true',
                          default=False, help='run in testing mode')
        parser.add_option('--config', dest='config_file', metavar='PATH',
                          help='path to monitor configuration file')
        parser.add_option('--name', dest='name', metavar='NAME',
                          help='plugin name')
        options, args = parser.parse_args()

        if args:
            parser.error('unknown arguments passed')

        if options.output_fd is None and not options.debug:
            parser.error('no output FD specified')

        if not options.interval and not options.debug:
            parser.error('no interval specified')

        if not options.config_file:
            parser.error('no config file location specified')
        if not os.path.isfile(options.config_file):
            parser.error('no such file: %s' % options.config_file)

        if not options.name and options.debug:
            # This is a hack :-P
            #pylint: disable-msg=W0212
            file_ = sys._getframe(3).f_code.co_filename
            options.name = os.path.splitext(os.path.basename(file_))[0]
            print 'Warning: no plugin name specified, fallback to "%s". ' \
                  'Use the --name option to set it.' % options.name
            print
        elif not options.name:
            options.error('no plugin name specified')

        if options.debug:
            if not options.interval:
                options.interval = 3

        return options.output_fd, options.interval, options.debug, \
                options.name, options.config_file

    def loop(self, fun):
        '''Loop a given function and C{produce} the return values'''
        def _loop():
            '''Actual looped function'''
            result = fun()
            for item in result:
                if len(item) == 3:
                    id_, type_, value = item
                    timestamp = long(time.time() * 1000)
                elif len(item) == 4:
                    if not isinstance(item[3], (int, long)):
                        raise TypeError('Timestamp is not of the correct ' \
                                        'type, are you using the old API?')
                    id_, type_, value, timestamp = item
                else:
                    raise ValueError(
                        'Producer function should return 4 or 5 items')

                self.produce(id_, type_, value, timestamp)

        while True:
            try:
                _loop()
                time.sleep(self.interval)
            except KeyboardInterrupt:
                break
            except:
                raise

    def get_config(self):
        '''Load configuration from disk'''
        cfg_fd = open(self.config_file, 'r')
        try:
            data = yaml.load(cfg_fd)
        finally:
            cfg_fd.close()

        try:
            cfg = data['plugins'][self.name]['config']
        except (TypeError, KeyError):
            cfg = dict()

        return cfg

    @staticmethod
    def get_nephos_config():
        '''Load global Nephos configuration from disk, if available'''
        try:
            from monitor import nephos_config as _nephos_config
            return _nephos_config.config
        except (ImportError, Exception), e: #pylint: disable-msg=W0703
            pass

    def install(self):
        '''Set up all required hooks (sys.excepthook for now)'''
        original_excepthook = sys.excepthook

        def excepthook(etype, evalue, etb):
            '''Handle a top-level exception

            This method will catch any InvalidPlatformException and signal the
            monitor accordingly.
            '''
            if not etype is InvalidPlatformException:
                return original_excepthook(etype, evalue, etb)

            if not self._packer or self.data_fd is None:
                return original_excepthook(etype, evalue, etb)

            message = ('invalid_platform_1', evalue.message, )
            serialized = json.dumps(message)

            self.data_fd.write(self._packer.pack(len(serialized)))
            self.data_fd.write(serialized)
            self.data_fd.flush()

            return original_excepthook(etype, evalue, etb)

        sys.excepthook = excepthook


__all__ = ['InvalidPlatformException', '__version__', ]
# Little shorthand
a = __all__.append

inst = Producer() #pylint: disable-msg=C0103
del Producer
inst.install()

produce = inst.produce #pylint: disable-msg=C0103
a('produce')
loop = inst.loop #pylint: disable-msg=C0103
a('loop')
notify = inst.notify #pylint: disable-msg=C0103
a('notify')
config = inst.get_config() #pylint: disable-msg=C0103
a('config')
name = inst.name #pylint: disable-msg=C0103
a('name')

nephos_config = inst.get_nephos_config()
if nephos_config is None:
    del nephos_config
else:
    a('nephos_config')

del inst
del a
del Struct
