# Copyright (C) 2011 Oracle. All rights reserved.
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, version 2.  This program is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the GNU
# General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 021110-1307, USA.

'''
Monitor plugin service
======================
This module implements a Soldjer service which allows plugins related to
monitoring and event notifications to be executed.

Plugins are simple Python scripts which use the 'produce' and 'loop' functions
defined in the 'monitor.producer' module and executed as-is. There's a sample
plugin in 'monitor.plugins.monitor'.

Plugin modules should reside in the path given to the 'Monitor' constructor, or
in 'monitor/plugins'.

Service structure
-----------------
The monitor service entry point is the 'Monitor' class, which is a
twisted.application.service.MultiService. When a 'Monitor' instance is created,
the plugin folder is scanned for plugins, and all plugins are loaded as a
service, and attached to the 'Monitor' multiservice.

Every plugin is managed by a 'MonitorPluginService' instance, which launches the
plugin in a subprocess (monitored by a 'MonitorPluginProto') and makes sure any
data produced by the plugin is sent to the 'Monitor'. It will also restart the
plugin whenever the runner process exits, unless the exit is planned.

Every 'MonitorPluginProto' contains a watchdog which makes sure a plugin
produces data within a certain grace period. If the plugin does not produce any
data, the underlying process is killed so it can be restarted.

The communication between plugin processes and the monitor happens through basic
pipes using a PString-like format, using JSON serialization.

The central 'Monitor' will rescan the plugin folder every N seconds. If the
module of a running plugin no longer exists, the plugin is killed. If the module
code of a running plugin is different than the code used in the current plugin
process, the plugin process is killed so the new code is executed when the
plugin process is restarted. When a new module is found, a new service is
created and started.

Consumers
---------
Consumers of published data should implement 'monitor.iconsumer.IConsumer' and
'twisted.application.service.IService'. Any consumer should be attached to a
monitor using the standard Twisted 'MultiService' hooking mechanism.

See 'monitor.consumers' for some helper classes.
'''
import os
import os.path
import sys
import time
import shutil
import struct
import signal
import inspect
import hashlib
import operator
import tempfile
import itertools

try: #pylint: disable-msg=R0801
    import json
except ImportError:
    import simplejson as json

import yaml

from twisted.internet import reactor, protocol, defer
from twisted.internet.error import ProcessExitedAlready
from twisted.internet.task import LoopingCall
from twisted.python import log
from twisted.python.reflect import namedAny
from twisted.protocols.basic import IntNStringReceiver
from twisted.application.service import Service, MultiService

# Disable some style-related PyLint messages, since this module is partially
# Twisted-style, which is PEP8-incompatible
#pylint: disable-msg=C0103

def _find_qbase_python_executable():
    '''Find the Python executable to use when spawning plugin processes'''
    if 'QBASE' in os.environ:
        return os.path.join(os.environ['QBASE'], 'bin', 'python')
    return sys.executable
QBASE_PYTHON = _find_qbase_python_executable()
del _find_qbase_python_executable

def _find_system_python_executable():
    '''Find the system Python executable path'''
    if 'QBASE' not in os.environ:
        return sys.executable

    path = os.environ.get('PATH', '')
    qbase = os.environ['QBASE']
    abs_qbase = os.path.abspath(qbase)

    paths = (p for p in path.split(os.pathsep) \
             if not p.startswith(qbase) \
                and not p.startswith(abs_qbase)
            )
    for path in paths:
        executable = os.path.join(path, 'python')
        if os.path.exists(executable):
            return executable

    return sys.executable
SYSTEM_PYTHON = _find_system_python_executable()
del _find_system_python_executable


def _create_fake_pythonpath():
    '''Create a minimal Python environment in a temporary directory using links

    The Nephos Monitor allows plugins to be executed using a system-wide
    installed Python executable. This might not be able to import the monitor,
    yaml and simplejson packages which is required for plugins.

    This function will create a base Python environment in a temporary directory
    using symlinks to these packages, which can be appended to the PYTHONPATH
    environment variable. If no symlink can be created, e.g. because the package
    is contained in an Egg file, the path (including the container file and the
    path inside it) will be returned in a tuple as second return value. The
    values inside this tuple should be appended to PYTHONPATH as well.

    @return: Fake Python environment location and Egg paths
    @rtype: str, tuple<str>
    '''
    try:
        import simplejson
    except ImportError:
        simplejson = None
    tmp = tempfile.gettempdir()
    base = os.path.join(tmp, 'nephos_monitor')
    if os.path.exists(base) and not os.path.isdir(base):
        raise Exception('File %s exists but is not a folder' % base)
    if not os.path.exists(base):
        os.mkdir(base)

    path = tempfile.mkdtemp('', 'monitor_fake_python_path', base)
    log.msg('Creating fake Python environment in %s' % path)
    extra_paths = list()

    dirname = os.path.dirname

    # Create all symlinks
    packages = (
        ('monitor', _create_fake_pythonpath),
        ('simplejson', simplejson),
        ('yaml', yaml),
    )
    for package_name, package_member in packages:
        if package_member is None:
            continue

        package_dir = dirname(inspect.getfile(package_member))
        if not os.path.isdir(package_dir):
            extra_paths.append(dirname(package_dir))
            continue

        dst = os.path.join(path, package_name)
        os.symlink(package_dir, dst)

    return path, tuple(extra_paths)


def calculate_hash(path):
    '''Calculate the SHA1 hash of a given file

    @param path: File to calculate hash
    @type path: str

    @return: SHA1 hash of the content of the given file
    @rtype: str
    '''
    fd = open(path)
    hasher = hashlib.sha1()
    bs = hasher.block_size

    while True:
        s = fd.read(bs)
        if not s:
            break
        hasher.update(s)
        # Maybe one day this will help the GC
        del s

    fd.close()

    return hasher.hexdigest()


class Logger(object): #pylint: disable-msg=R0903
    '''Twisted logging helper'''
    def __init__(self, name):
        '''Initialize a new Logger

        @param name: Name of the log system
        @type name: str
        '''
        self.name = name

    def msg(self, *args, **kwargs): #pylint: disable-msg=W0142
        '''Send a message to the Twisted log

        @see: L{twisted.python.log.msg}
        '''
        kwargs = kwargs.copy()
        kwargs['system'] = self.name
        log.msg(*args, **kwargs)


class _LengthLimitExceeded(Exception):
    '''Maximum string size exceeded

    This should never be raised, unless some specific condition in
    L{IntNStringReceiver} (of which L{MonitorPluginDataReceiver} is a subclass)
    is met.
    '''
    pass


class MonitorDataItem(object): #pylint: disable-msg=R0903
    '''Representation of one monitor data item'''
    __slots__ = ('_id', '_type', '_value', '_timestamp', )

    #pylint: disable-msg=R0913
    def __init__(self, id_, type_, value, timestamp):
        self._id = id_
        self._type = type_
        self._value = value
        self._timestamp = timestamp

    id = property(operator.attrgetter('_id'), doc='ID of the monitored item')
    type = property(operator.attrgetter('_type'), doc='Item type')
    value = property(operator.attrgetter('_value'), doc='Item value')
    timestamp = property(operator.attrgetter('_timestamp'),
                         doc='Item timestamp')

    def __str__(self):
        data = {
            'id': self._id,
            'type': self._type,
            'value': self._value,
            'timestamp': self._timestamp,
        }
        return str(data)


class EventItem(object): #pylint: disable-msg=R0903
    '''Representation of one event item'''
    __slots__ = ('_id', '_type', '_value', '_log', '_severity', '_timestamp', )

    #pylint: disable-msg=R0913
    def __init__(self, id_, type_, value, log_, severity, timestamp):
        self._id = id_
        self._type = type_
        self._value = value
        self._log = log_
        self._severity = severity
        self._timestamp = timestamp

    id = property(operator.attrgetter('_id'), doc='ID of the related item')
    type = property(operator.attrgetter('_type'), doc='Event type')
    value = property(operator.attrgetter('_value'), doc='Event value')
    log = property(operator.attrgetter('_log'), doc='Event log info')
    severity = property(operator.attrgetter('_severity'),
                        doc='Event severity')
    timestamp = property(operator.attrgetter('_timestamp'),
                         doc='Event timestamp')

    def __str__(self):
        data = {
            'id': self._id,
            'type': self._type,
            'value': self._value,
            'log': self._log,
            'severity': self._severity,
            'timestamp': self._timestamp,
        }
        return str(data)


class MonitorPluginDataReceiver(IntNStringReceiver):
    '''Protocol used to read plugin process output

    This implement a PString-like parser.
    '''
    # IntNStringReceiver prerequisites
    structFormat = '<L'
    prefixLength = struct.calcsize(structFormat)
    MAX_LENGTH = sys.maxint

    def __init__(self, callback_fun):
        '''Initialize a MonitorPluginDataReceiver

        @param callback_fun: Callable to be called when a complete string is
                             received
        @type callback_fun: callable
        '''
        self._callback_fun = callback_fun

    def stringReceived(self, msg):
        '''A complete string is received

        This method calls C{callable_fun} as provided in the constructor,
        passing through the received string.
        '''
        self._callback_fun(msg)

    def lengthLimitExceeded(self, _):
        '''
        Callback invoked when a length prefix greater than C{MAX_LENGTH} is
        received
        '''
        raise _LengthLimitExceeded


class MonitorPluginProto(protocol.ProcessProtocol):
    '''Protocol monitoring a plugin process'''
    MESSAGE_HANDLERS = dict()

    def __init__(self, gracetime, publish_fun, logger=None):
        '''Initialize a new MonitorPluginProto

        @param gracetime: Time (in seconds) in which a plugin *should* produce
                          data. If this timeframe is exceeded, the subprocess is
                          killed to be restarted.
        @type gracetime: number
        @param publish_fun: Function to which to pass plugin results
        @type publish_fun: callable
        '''
        assert callable(publish_fun)

        self._logger = logger or log

        self.gracetime = gracetime
        self.publish_fun = publish_fun
        self.process_exit = defer.Deferred()
        self.process_started = defer.Deferred()

        self._data_handler_helper = MonitorPluginDataReceiver(self._message)

        self.watchdog = None

    def connectionMade(self):
        '''A connection with the subprocess to monitor was established'''
        self.process_started.callback(self.transport.pid)
        self._logger.msg('Process %d started, launching watchdog with ' \
                        'gracetime %d' % (self.transport.pid, self.gracetime))

        def kill():
            '''Callback for the watchdog timer

            If a process didn't produce data within the grace period, it is
            killed.
            '''
            self.watchdog = None

            if not self.transport.pid:
                return

            self._logger.msg('Watchdog on process %d expired, killing ' \
                             'process' % self.transport.pid)
            try:
                self.transport.signalProcess(signal.SIGTERM)
            except ProcessExitedAlready: #pylint: disable-msg=W0704
                pass

        if self.gracetime > 0:
            #pylint: disable-msg=E1101
            self.watchdog = reactor.callLater(self.gracetime, kill)

    def childDataReceived(self, childFD, data):
        '''Some data was received from the child on FD C{childFD}'''
        if childFD == 1:
            if not data.strip():
                return
            self._logger.msg('[stdout@%d] %s' % \
                             (self.transport.pid, data.rstrip()))
        elif childFD == 2:
            if not data.strip():
                return
            self._logger.msg('[stderr@%d] %s' % \
                    (self.transport.pid, data.rstrip()))
        elif childFD == 3:
            if self.watchdog:
                # We got some data, reset our timeout
                self.watchdog.reset(self.gracetime)

            self._data_handler_helper.dataReceived(data)

    def processExited(self, _):
        '''Subprocess exited

        The subprocess quit, but there might be more data in buffers, which will
        be read before L{processEnded} is called. This is a good place to cancel
        our watchdog timer though.
        '''
        if self.watchdog:
            watchdog = self.watchdog
            self.watchdog = None
            watchdog.cancel()

    def processEnded(self, status):
        '''The process has quit and all buffers are read

        Time to launch a new subprocess. Call our C{process_exit} callback.
        '''
        self.process_exit.callback(status.value)

    def _message(self, msg):
        '''A full message was received from the plugin process

        The message string will be deserialized (using JSON notation) and passed
        to the publish function provided to the constructor.
        '''
        msg = json.loads(msg)
        message_type = msg[0]
        try:
            handler = self.MESSAGE_HANDLERS[message_type]
        except KeyError:
            raise TypeError('Unknown message type %s' % message_type)

        handler(self, *msg[1:]) #pylint: disable-msg=W0142

    def _handle_monitor_data_1(self, *args):
        '''Handle a 'monitor_data_1' type message'''
        id_, type_, value, timestamp = args
        item = MonitorDataItem(id_, type_, value, timestamp)
        self.publish_fun(item)
    MESSAGE_HANDLERS['monitor_data_1'] = _handle_monitor_data_1

    def _handle_event_1(self, *args):
        '''Handle a 'event_1' type message'''
        id_, type_, value, timestamp = args
        item = EventItem(id_, type_, None, log_, severity, timestamp)
        self.publish_fun(item)
    MESSAGE_HANDLERS['event_1'] = _handle_event_1

    def _handle_event_2(self, *args):
        '''Handle a 'event_2' type message'''
        id_, type_, value, log_, severity, timestamp = args
        item = EventItem(id_, type_, value, log_, severity, timestamp)
        self.publish_fun(item)
    MESSAGE_HANDLERS['event_2'] = _handle_event_2

    def _handle_invalid_platform_1(self, *args):
        '''Handle a 'invalid_platform_1' type message'''
        self.transport.invalid_platform = True
        msg, = args
        self._logger.msg('Plugin running on invalid platform: %s' % msg)
    MESSAGE_HANDLERS['invalid_platform_1'] = _handle_invalid_platform_1


class MonitorPluginService(Service): #pylint: disable-msg=R0902
    '''Service handling one monitor plugin process'''
    #pylint: disable-msg=R0913
    def __init__(self, name, path, interval, gracetime, publish_fun,
                 config_file, use_system_python):
        '''Initialize a new C{MonitorPluginService}

        @param name: Name of the plugin
        @type name: str
        @param path: Path of plugin module
        @type path: str
        @param interval: Interval between monitor status output
        @type interval: number
        @param gracetime: Time in which a plugin should produce data not to be
                          killed
        @type gracetime: number
        @param publish_fun: Function to call when monitoring data is received
        @type publish_fun: callable
        @param config_file: Config file location
        @type config_file: str
        @param use_system_python: Use system Python instead of Qbase Python
        @type use_system_python: bool
        '''
        assert os.path.isfile(path)
        assert callable(publish_fun)

        self._logger = Logger('monitor.plugin.%s' % name)

        self.path = path
        self.setName(name)
        self.interval = interval
        self.gracetime = gracetime
        self.publish_fun = publish_fun
        self.config_file = os.path.abspath(config_file)
        self.use_system_python = use_system_python

        self.transport = None
        self.stopping = False
        self._kill = False
        self.hash = None

    def startService(self):
        '''Start the service'''
        Service.startService(self)
        reactor.callWhenRunning(self._launch) #pylint: disable-msg=E1101

    def _launch(self): #pylint: disable-msg=R0912,R0914
        '''Launch the subprocess'''
        if self.stopping or self._kill:
            # We no longer *want* to be launched...
            return

        # Reset some stuff
        self.transport = None
        self.hash = None

        self._logger.msg('Starting plugin process %s' % self.path)

        self.hash = calculate_hash(self.path)

        # Create a handler protocol
        proto = MonitorPluginProto(self.gracetime, self.publish_fun,
                                   self._logger)

        def log_launch(pid):
            '''Log plugin process startup success'''
            self._logger.msg('Plugin process %s launched, PID %d' % \
                             (self.path, pid))
            return pid

        proto.process_started.addCallback(log_launch)

        def relaunch(arg):
            '''Relaunch a plugin process if it died of unnatural causes'''
            if not os.path.isfile(self.path):
                self._logger.msg('Plugin %s gone, not restarting' % self.path)
                self.transport = None
                self.disownServiceParent()
                return arg

            if self.stopping:
                self._logger.msg('Plugin process %s died, but service ' \
                                 'stopping: not restarting plugin process' % \
                                 self.path)
                return arg

            if self._kill:
                self._logger.msg('Plugin process %s died, but service was ' \
                                 'killed: not restarting plugin process' % \
                                 self.path)
                return arg

            if getattr(self.transport, 'invalid_platform', False):
                self._logger.msg('Plugin %s died, but running on an invalid ' \
                                 'platform, not restarting' % self.path)
                return arg

            self._logger.msg('Plugin process %s died, restarting' % self.path)
            self._launch()

            return arg

        fake_pythonpath = None
        extra_fake_pythonpath = None
        if self.use_system_python:
            # We need the extra 'fake' path entries which point into eggs and
            # can't be symlinked
            fake_pythonpath, extra_fake_pythonpath = _create_fake_pythonpath()

            def cleanup(arg):
                '''Cleanup the fake Python path after script exit'''
                self._logger.msg('Removing temporary Python path %s' % \
                                 fake_pythonpath)
                shutil.rmtree(fake_pythonpath, ignore_errors=True)
                return arg

            proto.process_exit.addBoth(cleanup)

        proto.process_exit.addCallback(relaunch)

        # Can't use Python 2.5 constructs...
        # executable = QBASE_PYTHON if not self.use_system_python \
        #                           else SYSTEM_PYTHON
        if not self.use_system_python:
            executable = QBASE_PYTHON
        else:
            executable = SYSTEM_PYTHON

        # -u: unbuffered output
        # output fd: used by L{monitor.producer.produce}
        args = [executable, '-u', self.path, '--output-fd', '3', '--interval',
                str(self.interval), '--config', self.config_file, '--name',
                self.name, ]
        env = {
            'PATH': os.environ.get('PATH', ''),
        }
        for key in ('HOME', 'USER', # System
                    'PYTHONPATH', 'PYTHONHOME', 'PYTHONUNBUFFERED', # Python
                    'QBASE'): # Nephos
            if key in os.environ:
                env[key] = os.environ[key]

        if self.use_system_python:
            pythonpath = env.get('PYTHONPATH', None)
            if not pythonpath:
                pythonpath = os.pathsep.join(itertools.chain(
                                (fake_pythonpath, ), extra_fake_pythonpath))
            else:
                pythonpath = os.pathsep.join(itertools.chain(
                                (pythonpath, fake_pythonpath, ),
                                extra_fake_pythonpath))

            env['PYTHONPATH'] = pythonpath

        childFDs = {
            1: 'r',
            2: 'r',
            3: 'r',
        }

        # This is ugly
        # It looks like even though we're using callWhenRunning everywhere
        # applicable, it is possible the reactor is not set up completely when
        # we reach this point (even when using addSystemEventTrigger('after',
        # 'startup'). This seems to happen on older versions of Twisted,
        # including the one shipped with the PyMonkey sandbox at this moment.
        # This code checks whether the SIGCHLD handler, which is required to be
        # registered before using spawnProcess, is actually registered. If not,
        # we call the signal install function ourself. Ugly, but works.
        try:
            import signal
        except ImportError:
            self._logger.msg('signal module not available')
        else:
            if hasattr(signal, 'SIGCHLD'):
                current_sigchld = signal.getsignal(signal.SIGCHLD)
                if not current_sigchld:
                    setup_fun = getattr(reactor, '_handleSignals',
                                        lambda: None)
                    setup_fun()


        #pylint: disable-msg=E1101
        self.transport = reactor.spawnProcess(proto, executable, args,
                                              env=env, childFDs=childFDs)

    def stopService(self):
        '''Stop the service

        @return: a L{Deferred} which is triggered when the service has
                 finished shutting down. If shutting down is immediate, C{None}
                 is returned.
        @rtype: L{Deferred}
        '''
        self.stopping = True

        if not self.transport or not self.transport.pid:
            # Nothing more to do
            return Service.stopService(self)

        self._logger.msg('Stopping plugin %s PID %d' % \
                         (self.path, self.transport.pid))

        try:
            self.transport.signalProcess(signal.SIGTERM)
        except ProcessExitedAlready: #pylint: disable-msg=W0704
            pass

        return Service.stopService(self)

    def kill(self, signal=signal.SIGTERM):
        '''Kill the plugin process

        @param signal: Signal to send to the subprocess, see
                       L{twisted.internet.interface.IReactorProcess.signalProcess}
        @type signal: str'''
        self._kill = True
        if self.transport and self.transport.pid:
            self._logger.msg('Killing process %d' % self.transport.pid)
            try:
                self.transport.signalProcess(signal)
            except ProcessExitedAlready: #pylint: disable-msg=W0704
                pass

    def restart(self):
        '''Restart the plugin process

        This simply kills the subprocess so it gets restarted automatically.
        '''
        if getattr(self.transport, 'invalid_platform', False):
            assert not self.transport.pid
            delattr(self.transport, 'invalid_platform')
            self._launch()
            return

        if self.transport and self.transport.pid:
            self._logger.msg('Restarting process %d' % self.transport.pid)
            try:
                self.transport.signalProcess(signal.SIGTERM)
            except ProcessExitedAlready: #pylint: disable-msg=W0704
                pass


class Monitor(MultiService):
    '''Main Monitor service

    This service discovers and launches monitor plugins, and handles monitoring
    data.
    '''
    def __init__(self, config):
        '''Initialize a new monitor

        A plugin folder can be provided. If none is given, a default path
        relative to this module ('../plugins') is calculated.

        @param plugin_base: Path to folder containing all plugin scripts
        @type plugin_base: str
        '''
        MultiService.__init__(self)

        self._logger = Logger('monitor')
        self._logger.msg('Initializing monitor using config in %s' % config)

        self._config_file = config
        self._config = self._parse_config(config)

        self._reload_task = None

        plugin_path = self._config['config']['plugin_path']
        if plugin_path[0] != '/':
            plugin_path = os.path.abspath(os.path.join(
                            os.path.dirname(__file__), plugin_path))
        assert os.path.isdir(plugin_path)
        self._logger.msg('Plugin path is %s' % plugin_path)
        self.plugin_path = plugin_path

        self.plugins = MultiService()
        self.plugins.setName('plugins')
        self.plugins.setServiceParent(self)
        self.consumers = MultiService()
        self.consumers.setName('consumers')
        self.consumers.setServiceParent(self)

        self._load_consumers()

    def startService(self):
        '''Start the service'''
        ret = MultiService.startService(self) #pylint: disable-msg=E1111

        self._reload_task = LoopingCall(self._reload, self.plugin_path)
        #pylint: disable-msg=E1101
        reload_interval = self._config['config']['reload_interval']
        reactor.callWhenRunning(self._reload_task.start, reload_interval)

        return ret

    def stopService(self):
        '''Stop the service'''
        if self._reload_task:
            self._reload_task.stop()
        return MultiService.stopService(self)

    def publish(self, data):
        '''Handle some monitoring data provided by a plugin process

        This should be implemented in a subclass.

        @param data: Data received
        @type data: object
        '''
        for service in self.consumers.services:
            service.publish(data)

    def _reload(self, plugin_base):
        '''Reload all plugins

        This method will list all plugin scripts in C{base_path}, kill any
        running plugins which no longer exist, spawn new ones and reload
        existing ones of which the code was changed on-disk.

        @param plugin_base: Folder containing plugin scripts
        @type plugin_base: str
        '''
        self._logger.msg('Rescanning all plugins')

        # Reload configuration (will only apply on new plugins!)
        self._config = self._parse_config(self._config_file)

        plugin_files = set(self._find_plugins(plugin_base))
        current_plugins = set(s.path for s in self.plugins.services)

        # Kill all plugins whose files are gone
        to_kill = current_plugins.difference(plugin_files)
        for path in to_kill:
            # Find the service (MonitorPluginService that is) which manages this
            # plugin
            matches = [service for service in self.plugins.services \
                        if service.path == path]
            # We should only find 0 or 1 result
            assert len(matches) < 2

            if not matches:
                continue

            self._logger.msg('File %s gone, killing service' % path)
            self.plugins.removeService(matches[0])

        # Launch new plugins
        new_plugins = plugin_files.difference(current_plugins)
        for path in new_plugins:
            name = os.path.splitext(os.path.basename(path))[0]
            try:
                interval = self._config['plugins'][name]['interval']
            except (TypeError, KeyError), ex:
                interval = self._config['config']['default_interval']
            try:
                gracetime = self._config['plugins'][name]['grace']
            except (TypeError, KeyError), ex:
                gracetime = self._config['config']['default_grace']
            try:
                use_system_python = \
                    self._config['plugins'][name]['use_system_python']
            except (TypeError, KeyError), ex:
                use_system_python = False

            service = MonitorPluginService(name, path, interval, gracetime,
                                           self.publish, self._config_file,
                                           use_system_python)
            service.setServiceParent(self.plugins)

        # Kill/restart all plugins whose source hash has been changed
        existing_plugins = plugin_files.intersection(current_plugins)
        for path in existing_plugins:
            matches = [service for service in self.plugins.services \
                        if service.path == path]
            assert len(matches) < 2

            if not matches:
                continue

            service = matches[0]

            new_hash = calculate_hash(path)
            if getattr(service, 'hash', None) in (new_hash, None, ):
                continue

            self._logger.msg('Hash of %s changed, restarting existing ' \
                             'plugin process' % path)
            service.restart()

    @staticmethod
    def _find_plugins(plugin_base):
        '''List all plugins in C{plugin_base}'''
        for name in os.listdir(plugin_base):
            name = os.path.join(plugin_base, name)
            if not name.endswith('.py'):
                continue
            if not os.path.isfile(name):
                continue

            yield name

    @staticmethod
    def _parse_config(path):
        '''Parse config file in given path

        @param path: Config file location
        @type path: str
        '''
        fd = open(path, 'r')
        try:
            config = yaml.load(fd)
        finally:
            fd.close()
        return config

    def _load_consumers(self):
        '''Load and set up all configured consumers'''
        self._logger.msg('Loading all consumers')
        try:
            config = self._config['consumers']
        except KeyError:
            self._logger.msg('No consumers defined')
            return

        for name, conf in config.iteritems():
            self._logger.msg('Setting up consumer %s' % name)
            handler = conf['handler']
            args = conf.get('args', dict())
            handler_class = namedAny(handler)
            svc = handler_class(**args) #pylint: disable-msg=W0142
            svc.setName(name)
            svc.setServiceParent(self.consumers)
